/*
**   UltimateIRCd - Ultimate Internet Relay Chat Daemon, src/s_hide.c
**   Copyright (C) Roger Yerramsetti
**
**   Rewritten for UltimateIRCd by ShadowRealm Creations.
**   Copyright (C) 1998-2000 Infomedia Inc.
**
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**
** Rewritten from AustNet's VirtualWorld Code.
** Virtual World written by Roger 'RogerY' Yerramsetti <rogery@austnet.org>
** Rewritten and used with permission from the Author.
**
*/

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"


/* Hidden host code below */

#define MAXVIRTSIZE     (3 + 5 + 1)
#define HASHVAL_TOTAL   40015
#define HASHVAL_PARTIAL 315

extern  aClient me;
extern  int     seed;
int     match(char *, char *), find_exception(char *);

extern unsigned char tolowertab[];

int str2array(char **pparv, char *string, char *delim)
{
    char *tok;
    int pparc=0;

    tok=(char *)strtok(string,delim);
    while(tok != NULL)
    {
        pparv[pparc++]=tok;
        tok=(char *)strtok(NULL,delim);
    }

    return pparc;
}


void truncstring(char *stringvar, int firstlast, int amount){
   if (firstlast)
   {
    stringvar+=amount;
    *stringvar=0;
    stringvar-=amount;
   }
    else
   {
    stringvar+=strlen(stringvar);
    stringvar-=amount;
   }
}

int Maskchecksum (char *data, int len)
{
        int                     i;
        int                     j;

        j=0;
        for (i=0 ; i<len ; i++)
        {
          j += *data++ * (i < 16 ? (i+1)*(i+1) : i*(i-15));

        }

        return (j+1000)%0xffff;
}

unsigned z(char *s)
{
    unsigned int i;
        i = Maskchecksum (s, strlen(s));
    return i;
}

/* Regular user host */
void    make_virthost(char *curr, char *new)
{
    unsigned hash_total = z(curr) % HASHVAL_TOTAL, hash_partial;
    char     destroy[HOSTLEN+1], *parv[HOSTLEN+1], tmpnew[HOSTLEN], *ptr;
    int      parc=0, len=0, overflow=0, exception=0;

    strncpyzt(destroy, curr, HOSTLEN);

    len = strlen(destroy);

    if ((len + MAXVIRTSIZE) > HOSTLEN)
    {
        overflow = (len + MAXVIRTSIZE) - HOSTLEN;

        ptr = &destroy[overflow];
    }
    else
        ptr = &destroy[0];

    parc = str2array(parv, ptr, ".");


    if (parc == 8)
    {
	if ((parv[parc-2] == "home") && (parv[parc-1] == "com"))
	{
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s.%s",
        hidden_host, hash_total, parv[parc-4], parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
	}
	else
	{
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s.%s.%s.%s.%s",
        hidden_host, hash_total, parv[parc-7], parv[parc-6], parv[parc-5], parv[parc-4], parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
     	}
    }

    if (parc == 7)
    {
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s.%s.%s.%s",
        hidden_host, hash_total, parv[parc-6], parv[parc-5], parv[parc-4], parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }
    if (parc == 6)
    {
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s.%s.%s",
        hidden_host, hash_total, parv[parc-5], parv[parc-4], parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }

    if (parc == 5)
    {
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s.%s",
        hidden_host, hash_total, parv[parc-4], parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }

    if (parc == 4)
    {
        len = strlen(parv[3]);

        if (strchr("0123456789", parv[3][len-1]))
        {
            hash_partial = (z(parv[0])+z(parv[1])+z(parv[2])) % HASHVAL_PARTIAL;
            ircsprintf(tmpnew, "%s.%s.%u.%u",
                parv[0], parv[1], hash_partial, hash_total);

            strncpyzt(new, tmpnew, HOSTLEN);
            return;
        }

	else {
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s",
        hidden_host, hash_total, parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
	}
    }



    if (parc == 3)
    {
        ircsprintf(tmpnew, "%s-%u.%s.%s",
        hidden_host, hash_total, parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }


    if (parc == 2)
    {
        ircsprintf(tmpnew, "%s-%u.%s.%s",
            hidden_host, hash_total, parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }

    len = strlen(parv[parc-1]);

    if ((len==2) && (find_exception(parv[parc-1])))
        exception = 1;

    if ((len == 2) && (exception == 0))
    {
        ircsprintf(tmpnew, "%s-%u.%s.%s.%s",
            hidden_host, hash_total, parv[parc-3], parv[parc-2], parv[parc-1]);

        strncpyzt(new, tmpnew, HOSTLEN);
        return;
    }
    ircsprintf(tmpnew, "%s-%u.%s.%s", hidden_host, hash_total, parv[parc-2], parv[parc-1]);
    strncpyzt(new, tmpnew, HOSTLEN);
    return;
}

/* NetAdmin host */
void	make_netadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", netadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}


/* IRCAdmin host */
void	make_ircadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", ircadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* ServicesAdmin host */
void	make_servicesadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", servicesadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* ServicesRoot host */
void	make_servicesroothost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", servicesroot_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* TechAdmin host */
void	make_techadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", techadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* ServAdmin host */
void	make_servadminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", servadmin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* Admin host */
void	make_adminhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", admin_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* ServicesOper host */
void	make_servicesoperhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", servicesoper_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* Oper host */
void	make_operhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", oper_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* LocalOper host */
void	make_locophost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", locop_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* NetMon host */
void	make_netmonhost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", netmon_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

/* SBot host */
void	make_sbothost(char *new)
{
    char     tmpnew[HOSTLEN];

        ircsprintf(tmpnew, "%s", sbot_host);
        strncpyzt(new, tmpnew, HOSTLEN);
        return;
}

